/**
 * This class represents is the main class to start an agent
 *
 * @author  Klaus Dorer
 * living systems AG
 */
package com.ls.research.robocup;

import com.ls.research.robocup.server.ServerRepresentation;
import com.ls.research.robocup.perception.PerceptionInterface;
import com.ls.research.robocup.perception.PerceptionParser;
import com.ls.research.robocup.perception.VisualInfo;
import com.ls.research.robocup.perception.AudibelInfo;
import com.ls.research.robocup.perception.PropriozeptiveInfo;

import java.util.HashMap;
import java.util.Hashtable;
import java.util.Map;

public class ExampleAgent implements PerceptionInterface
{
	private String					m_hostName;
	private int 					m_port;
	private int						m_senseCount;
	private String					m_team;
	private ServerRepresentation 	m_server;
	private PerceptionParser		m_perception;

	/**
	 * Command line format:
	 *
	 * SoccerAgent [-parameter value]
	 *
	 * Parameters:
	 *
	 *	host (default "localhost")
	 *		The host name can either be a machine name, such as "java.sun.com"
	 *		or a string representing its IP address, such as "127.0.0.1"
	 *
	 *	port (default 6000)
	 *		Port number for communication with server
	 *
	 *	team (default living-systems)
	 *		Team name. This name can not contain spaces.
	 *
	 */
	public static void main(String[] args)
	{
		String					hostName 	= new String("127.0.0.1");
		int						port 		= 6000;
		String					team 		= new String("living");

		try
		{
			// First look for parameters
			for( int c = 0 ; c < args.length ; c += 2 )
			{
				if( args[c].compareTo("-host") == 0 )
				{
					hostName = args[c + 1];
				}

				else if( args[c].compareTo("-port") == 0 )
				{
					port = Integer.parseInt(args[c + 1]);
				}
				else if( args[c].compareTo("-team") == 0 )
				{
					team = args[c + 1];
				}
				else
				{
					throw new Exception();
				}
			}

			ExampleAgent agent = new ExampleAgent(hostName, port, team);
			agent.start();
		}
		catch(Exception e)
		{
			System.err.println("");
			System.err.println("USAGE: java SoccerAgent [-parameter value]");
			System.err.println("");
			System.err.println("    Parameters  value        default");
			System.err.println("   ------------------------------------");
			System.err.println("    host        host_name    localhost");
			System.err.println("    port        port_number  6000");
			System.err.println("    team        team_name    living-systems");
			System.err.println("");
			System.err.println("    Example:");
			System.err.println("      java SoccerAgent -host www.host.com -port 6000 -team living-systems");
			System.err.println("    or");
			System.err.println("      java SoccerAgent -host 127.0.0.1");
			return;
		}
	}

	/**
	 * Constructor creating an agent
	 *
	 * @param host		hostname of the server
	 * @param port		port used by the server
	 * @param team		team name the agent should join
	 */
	public ExampleAgent(String host, int port, String team)
	{
		m_hostName 		= host;
		m_port 			= port;
		m_team 			= team;
		m_senseCount 	= 0;
		m_server		= null;
		m_perception	= null;
	}

	public void start()
	{
		try
		{
			// create server
			System.out.println("creating server representation...");
			m_perception = new PerceptionParser(this);
			m_server = new ServerRepresentation(m_perception, m_hostName, m_port);

			// connect to server
			System.out.println("connecting to server...");
			m_server.connectToServer(m_team, "version " + RobocupConstants.SERVER_VERSION_USED + ".00", false);

			System.err.println("playing...");
			try{
				Thread.sleep(5000);
			} catch(Exception e) {}

			System.err.println("disconnecting...");
			m_server.disconnectFromServer();

		} catch (Exception e) {
			System.out.println("start: " + e.getMessage());
		}
	}

	/**
	 * Called when any other that the previos messages arrived such as init or error
	 *
	 * @param message	unparsed message of the server
	 */
	public void onMessage(String message)
	{
		System.out.println(message);
	}


	/**
	 * Called when new see information arrived from the server
	 *
	 * @param visual	parsed see message of the server
	 */
	public void onSee(VisualInfo visual)
	{
	}

	/**
	 * Called when new hear information arrived from the server
	 *
	 * @param audibel	parsed hear message of the server
	 */
	public void onHear(AudibelInfo audibel)
	{
	}

	/**
	 * Called when new sense body information arrived from the server
	 *
	 * @param propriozeption	parsed sense body message of the server
	 */
	public void onSenseBody(PropriozeptiveInfo propriozeption)
	{
		doSomething();
	}

	/**
	 * Called when init information arrived from the server
	 *
	 * @param side		side on which the player plays l|r
	 * @param number	number of the player
	 * @param team		teamname the player joins
	 */
	public void onInit(char side, int number, String team)
	{
		System.out.println("init: " + side + " " + number + " " + team);
	}

	/**
	 * Called when the server parameter information is received at the start of the game
	 *
	 * @param parameters	parameters of the server stored in a hashmap
	 */
	public void onServerParam(HashMap parameters)
	{
	}

	/**
	 * Called when the parameter information of the player is received
	 *
	 * @param parameters	parameters of the player stored in a hashmap
	 */
	public void onPlayerParam(HashMap parameters)
	{
	}

	/**
	 * Called when the player type information of the player is received
	 *
	 * @param parameters	parameters of the player type stored in a hashmap
	 */
	public void onPlayerType(HashMap parameters)
	{
	}

	/**
	 * Called when a player changed type
	 *
	 * @param number	number of the player that changed
	 * @param type		the new type of the player, if the player is of our own team
	 *					ObjectInfo.NOT_EXISTING if the player is of the other team
	 */
	public void onChangePlayer(int number, int type)
	{
	}

	/**
	 * Called when a player received score information in reaction to a score command
	 *
	 * @param time		the time of the server
	 * @param left		score of the left team at time
	 * @param right		score of the right team at time
	 */
	public void onScore(int time, int left, int right)
	{
	}

	/**
	 * @return the name of the team the agent belongs to
	 */
	public String getTeam()
	{
		return m_team;
	}

	/** Called whenever the agent should decide to do the next step.
	 *  Expand this method to get intelligent agents :-)
	 */
	 protected void doSomething()
	 {
		if (m_senseCount++ < 15)
			m_server.turn(30);
		else
			m_server.dash(100);
	}
}
