/**
 * This class is parsing server perception information and notifying the agent
 *
 * @author  Klaus Dorer
 */
package com.ls.research.robocup.perception;

import com.ls.research.robocup.server.SoccerPerceptionInterface;

import java.util.StringTokenizer;
import java.util.HashMap;

public class PerceptionParser implements SoccerPerceptionInterface
{
	private PerceptionInterface 	m_agent;
	private	VisualInfo 				m_visual;
	private PropriozeptiveInfo 		m_proprio;

	/**
	 * Constructor creating the parser
	 *
	 * @param agent		reference to a class that implements the perception interface
	 */
	public PerceptionParser(PerceptionInterface agent)
	{
		m_agent = agent;
		m_visual = new VisualInfo();
		m_proprio = new PropriozeptiveInfo();
	}

	/**
	 * Called when new see information arrived from the server.
	 * Parses the info into a VisualObject and informs the agent.
	 *
	 * @param message	unparsed see message of the server
	 */
	public void onSee(String message)
	{
		m_visual.parse(message);
		m_agent.onSee(m_visual);
	}

	/**
	 * Called when new hear information arrived from the server
	 *
	 * @param message	unparsed hear message of the server
	 */
	public void onHear(String message)
	{
		AudibelInfo audio = new AudibelInfo(message);
		audio.parse();
		m_agent.onHear(audio);
	}

	/**
	 * Called when new sense body information arrived from the server
	 *
	 * @param message	unparsed sense body message of the server
	 */
	public void onSenseBody(String message)
	{
		m_proprio.parse(message);
		m_agent.onSenseBody(m_proprio);
	}

	/**
	 * Called when init information arrived from the server
	 *
	 * @param message	unparsed init message of the server
	 */
	public void onInit(String message)
	{
		StringTokenizer	tokenizer = new StringTokenizer(message, "() ");

		try {
			// (init l 9 living)
			tokenizer.nextToken();
			char side   = tokenizer.nextToken().charAt(0);
			int	 number = Integer.parseInt(tokenizer.nextToken());

			m_agent.onInit(	side, number, tokenizer.nextToken());

		} catch (Exception e) {
			System.out.println("Error parsing init information");
			while( tokenizer.hasMoreTokens() )
				System.out.print(tokenizer.nextToken());
			System.out.println("");
		}
	}

	/**
	 * Called when the server parameter information is received at the start of the game
	 *
	 * @param message	unparsed parameter information of the server
	 */
	public void onServerParam(String message)
	{
		HashMap	parameters;
		parameters = parseParameters(message, PerceptionInterface.server_param);
		m_agent.onServerParam(parameters);
	}

	/**
	 * Called when the parameter information of the player is received
	 *
	 * @param message	unparsed parameter information of the player
	 */
	public void onPlayerParam(String message)
	{
		HashMap	parameters;
		parameters = parseParameters(message, PerceptionInterface.player_param);
		m_agent.onPlayerParam(parameters);
	}


	/**
	 * Called when the player type information of players is received
	 *
	 * @param message	unparsed player type information
	 */
	public void onPlayerType(String message)
	{
		HashMap	parameters;
		parameters = parseParameters(message, PerceptionInterface.player_type);
		m_agent.onPlayerType(parameters);
	}

	/**
	 * Called when the a player changed type
	 *
	 * @param message	unparsed change type information of the server
	 */
	public void onChangePlayer(String message)
	{
		StringTokenizer	tokenizer = new StringTokenizer(message, ") ");

		try {
			// (change_player_type l 9)
			tokenizer.nextToken();
			int	 number = Integer.parseInt(tokenizer.nextToken());
			int	 type;
			if (tokenizer.hasMoreTokens())
				type = Integer.parseInt(tokenizer.nextToken());
			else
				type = ObjectInfo.NOT_EXISTING;

			m_agent.onChangePlayer(number, type);

		} catch (Exception e) {
			System.out.println("Error parsing change player information: " + message);
		}
	}

	/**
	 * Called when any other that the previos messages arrived such as init or error
	 *
	 * @param message	unparsed message of the server
	 */
	public void onMessage(String message)
	{
		System.out.println(message);
	}

	/**
	 * Called when the player receives the score of the game after sending score action
	 *
	 * @param message	unparsed score information of the server
	 */
	public void onScore(String message)
	{
		StringTokenizer	tokenizer = new StringTokenizer(message, ") ");

		try {
			// (score time 3 1)
			tokenizer.nextToken();
			int	 time  = Integer.parseInt(tokenizer.nextToken());
			int	 left  = Integer.parseInt(tokenizer.nextToken());
			int	 right = Integer.parseInt(tokenizer.nextToken());

			m_agent.onScore(time, left, right);

		} catch (Exception e) {
			System.out.println("Error parsing change player information: " + message);
		}
	}

	/**
	 * Called when any other that the previos messages arrived such as init or error
	 *
	 * @param message	unparsed message of the server
	 */
	private HashMap parseParameters(String message, String[] names)
	{
		HashMap 		parameters	= new HashMap();
		StringTokenizer	tokenizer = new StringTokenizer(message.trim(), ") ");
		Float			parameterValue;

		try
		{
			// overread (player_param
			tokenizer.nextToken();

			// read all parameters
			for (int i = 0; i < names.length; i++)
			{
				parameterValue = new Float(tokenizer.nextToken());
				parameters.put(names[i], parameterValue);
			}

			if (tokenizer.hasMoreTokens())
			{
				System.out.println("Server or player parameters left over while parsing\n");
				while( tokenizer.hasMoreTokens() )
					System.out.print(tokenizer.nextToken());
				System.out.println("\n Message: " + message);
			}
		}
		catch (Exception e)
		{
			while( tokenizer.hasMoreTokens() )
				System.out.print(tokenizer.nextToken());
			System.out.println("");
			System.out.println("Invalid parameters string: " + e.getMessage() + "\n" + message);
		}

		return parameters;
	}
}