/****************************************************************************
 *	This is the class representing all visual information
 *	Members are public since this are only data containers
 ***************************************************************************/
package com.ls.research.robocup.perception;

import java.util.StringTokenizer;
import java.util.Vector;
import java.util.Iterator;
import java.util.NoSuchElementException;

public class VisualInfo
{
	public static final int NO_OF_PLAYERS_PER_TEAM 	= 11;
	public static final int NO_OF_PLAYERS			= 22;

	public static final int NO_OF_FLAGS		= 50;
	public static final int CENTER_MARK		= 0;
	public static final int CENTER_TOP		= 1;
	public static final int CENTER_BOTTOM	= 2;
	public static final int GOAL_POST_L		= 3;
	public static final int GOAL_POST_R		= 5;
	public static final int FLAG_OUTSIDE_TL	= 7;
	public static final int FLAG_OUTSIDE_TC	= 12;
	public static final int FLAG_OUTSIDE_TR	= 13;
	public static final int FLAG_OUTSIDE_BL	= 18;
	public static final int FLAG_OUTSIDE_BC	= 23;
	public static final int FLAG_OUTSIDE_BR	= 24;
	public static final int FLAG_OUTSIDE_LT	= 29;
	public static final int FLAG_OUTSIDE_LC	= 32;
	public static final int FLAG_OUTSIDE_LB	= 33;
	public static final int FLAG_OUTSIDE_RT	= 36;
	public static final int FLAG_OUTSIDE_RC	= 39;
	public static final int FLAG_OUTSIDE_RB	= 40;
	public static final int FLAG_PENALTY_L	= 43;
	public static final int FLAG_PENALTY_R	= 46;
	public static final int TRASH_FLAG		= 49;

	public static final double[] FLAG_POSITIONS_X  =
				{ 0.0, 0.0, 0.0, -52.5, -52.5, 52.5, 52.5
				, -10.0, -20.0, -30.0, -40.0, -50.0, 0.0
				, 10.0, 20.0, 30.0, 40.0, 50.0
				, -10.0, -20.0, -30.0, -40.0, -50.0, 0.0
				, 10.0, 20.0, 30.0, 40.0, 50.0
				, -57.5, -57.5, -57.5, -57.5
				, -57.5, -57.5, -57.5
				, 57.5, 57.5, 57.5, 57.5
				, 57.5, 57.5, 57.5
				, -36.0, -36.0, -36.0, 36.0, 36.0, 36.0, 0.0 };

	public static final double[] FLAG_POSITIONS_Y  =
				{ 0.0, -34.0, 34.0, -7.01, 7.01, -7.01, 7.01
				, -39.0, -39.0, -39.0, -39.0, -39.0, -39.0
				, -39.0, -39.0, -39.0, -39.0, -39.0
				, 39.0, 39.0, 39.0, 39.0, 39.0, 39.0
				, 39.0, 39.0, 39.0, 39.0, 39.0
				, -10.0, -20.0, -30.0, 0.0
				, 10.0, 20.0, 30.0
				, -10.0, -20.0, -30.0, 0.0
				, 10.0, 20.0, 30.0
				, -20.16, 0.0, 20.16, -20.16, 0.0, 20.16, 0.0 };

	public	int				m_time;
	public	ObjectInfo		m_ball;
	public	ObjectInfo		m_leftGoal;
	public	ObjectInfo		m_rightGoal;
	public	ObjectInfo[]	m_flags;
//	public	ObjectInfo[]	m_lines;
	public	ObjectInfo	    m_dummyLine;
	public	PlayerInfo[]	m_players;
	public	int				m_playerCount;

	private	StringTokenizer	m_tokenizer;
	private String			m_token;
	private Vector          m_visibleObjects;

	/**
	 * Constructor creating all visible objects. This should be done only once.
	 */
	public VisualInfo()
	{
		m_visibleObjects = new Vector(NO_OF_PLAYERS + NO_OF_FLAGS + 3);
		m_time		= 0;

		m_ball		= new ObjectInfo();
		m_visibleObjects.add(m_ball);

		m_leftGoal	= new ObjectInfo();
		m_visibleObjects.add(m_leftGoal);

		m_rightGoal	= new ObjectInfo();
		m_visibleObjects.add(m_rightGoal);

		m_flags 	= new ObjectInfo[NO_OF_FLAGS];
		for (int i = 0; i < NO_OF_FLAGS; i++)
		{
			m_flags[i] = new ObjectInfo();
			m_visibleObjects.add(m_flags[i]);
		}

//		m_lines 	= new ObjectInfo[NO_OF_LINES];
		m_dummyLine	= new ObjectInfo();
		m_players	= new PlayerInfo[NO_OF_PLAYERS];
		for (int i = 0; i < NO_OF_PLAYERS; i++)
		{
			m_players[i] = new PlayerInfo();
			m_visibleObjects.add(m_players[i]);
		}
		m_playerCount = 0;
	}

	/**
	 * initializes all visible objects
	 */
	public void initialize()
	{
		Iterator objects = m_visibleObjects.iterator();
		while (objects.hasNext())
		{
			((ObjectInfo) objects.next()).initialize();
		}

		m_playerCount = 0;
	}

	/**
	 * Parses the visual string into java objects
	 *
	 * @param message	message to be parsed
	 */
	public void parse(String message)
	{
		ObjectInfo	objInfo;

		try
		{
			initialize();

			message.trim();
			m_tokenizer = new StringTokenizer(message,"() ", true);

			// First is '('
			m_tokenizer.nextToken();
			// then we should find see
			m_tokenizer.nextToken();
			// now ' ' followed by time and ' '
			m_tokenizer.nextToken();
			m_time = Integer.parseInt( m_tokenizer.nextToken() );
			m_tokenizer.nextToken();

			// then we should find '('
			m_token = m_tokenizer.nextToken();

			while(m_token.equals("("))
			{
				// we have another object so go on and get its name.
				objInfo = createNewObject();
				objInfo.m_isVisible = true;
				if (findEnd())
					continue;

				// get distance
				m_token = m_tokenizer.nextToken();
//				System.out.println("dist: " + m_token);
				objInfo.m_distance = Float.valueOf(m_token).floatValue();
				if (findEnd())
					continue;

				// get direction
				m_token = m_tokenizer.nextToken();
//				System.out.println("dir: " + m_token);
				objInfo.m_direction = Integer.valueOf(m_token).intValue();
				if (findEnd())
					continue;

				// get distance change
				m_token = m_tokenizer.nextToken();
//				System.out.println("distch: " + m_token);
				objInfo.m_distChange = Double.valueOf(m_token).floatValue();
				if (findEnd())
					continue;

				// get direction change
				m_token = m_tokenizer.nextToken();
//				System.out.println("dirch: " + m_token);
				float dirChange = Double.valueOf(m_token).floatValue();
				objInfo.m_dirChange = (int) dirChange;
				m_token = m_tokenizer.nextToken();
				m_token = m_tokenizer.nextToken();
				if(m_token.equals(" "))
					m_token = m_tokenizer.nextToken();
			}
		}
		catch(Exception e)
		{
			System.out.println("Error parsing see information");
			System.out.println(e);
			while( m_tokenizer.hasMoreTokens() )
				System.out.print(m_tokenizer.nextToken());

			System.out.println("\n original message: \n" + message);
		}
	}

	/**
	 * Parses and creates a specific object
	 *
	 * @return	object representing a visible object
	 */
	private ObjectInfo createNewObject()
	{
		ObjectInfo		objInfo = null;

		// bypass first "("
		m_tokenizer.nextToken();
		m_token = m_tokenizer.nextToken();

		if(m_token.startsWith("p")) {
			// we have a player
			String 	team 	= new String();
			int 	number 	= 0;
			boolean	goalie	= false;
			m_token = m_tokenizer.nextToken();	// space
			if(m_token.equals(" ")) {
				// we have team and number information
				team = m_tokenizer.nextToken();		// teamname
				m_token = m_tokenizer.nextToken();	// space
				if (m_token.equals(" ")) {
					number = Integer.parseInt(m_tokenizer.nextToken());	// uniform number
					m_token = m_tokenizer.nextToken();	// space
					if (m_token.equals("goalie")) {
						goalie = true;
					}
				}
			}
			objInfo = m_players[m_playerCount++];
			((PlayerInfo)objInfo).m_teamName 	= team;
			((PlayerInfo)objInfo).m_number 		= number;
			((PlayerInfo)objInfo).m_isGoalie 	= goalie;

		} else if(m_token.startsWith("f")) {
			int	 iFlag = TRASH_FLAG;
			char tag1  = 0;
			char tag2  = 0;
			char tag3  = 0;
			int	 posid = 0;

			m_token = m_tokenizer.nextToken();	// space
			m_token = m_tokenizer.nextToken();	// p or [l|c|r] or [t|b|r|l]
			tag1 = m_token.charAt(0);
			m_token = m_tokenizer.nextToken();	// space or ) if (flag c)

			if (!m_token.equals(")")) {
				// we do not have the (flag c)
				m_token = m_tokenizer.nextToken();
				tag2 = m_token.charAt(0);
				m_token = m_tokenizer.nextToken();	// space or )
				if (!m_token.equals(")")) {
					m_token = m_tokenizer.nextToken();
					tag3 = m_token.charAt(0);
				}
			}

			switch (tag1)
			{
				case 'c':
					// it is a center flag
					if (tag2 == 't')
						iFlag = CENTER_TOP;
					else if (tag2 == 'b')
						iFlag = CENTER_BOTTOM;
					else
						iFlag = CENTER_MARK;
					break;

				case 'g':
					// it is a goal flag
					if (tag2 == 'l')
					{
						if (tag3 == 't')
							iFlag = GOAL_POST_L;
						else
							iFlag = GOAL_POST_L + 1;
					}
					else
					{
						if (tag3 == 't')
							iFlag = GOAL_POST_R;
						else
							iFlag = GOAL_POST_R + 1;
					}
					break;

				case 'p':
					// it is a flag of the penalty area
					if (tag2 == 'l')
					{
						if (tag3 == 't')
							iFlag = FLAG_PENALTY_L;
						else if (tag3 == 'c')
							iFlag = FLAG_PENALTY_L + 1;
						else
							iFlag = FLAG_PENALTY_L + 2;
					}
					else
					{
						if (tag3 == 't')
							iFlag = FLAG_PENALTY_R;
						else if (tag3 == 'c')
							iFlag = FLAG_PENALTY_R + 1;
						else
							iFlag = FLAG_PENALTY_R + 2;
					}
					break;

				case 'l':
					// it is one of the left flags of the field
					if (tag2 == '0')
						// it is the left center flag
						iFlag = FLAG_OUTSIDE_LC;
					else
					{
						// find the proper flag index
						switch (tag3)
						{
							case '5': posid++;
							case '4': posid++;
							case '3': posid++;
							case '2': posid++;
						}
						if (tag2 == 't')
							// it is a left top outside flag
							iFlag = FLAG_OUTSIDE_LT + posid;
						else
							// it is a left bottom outside flag
							iFlag = FLAG_OUTSIDE_LB + posid;
					}
					break;

				case 'r':
					// it is one of the right flags of the field
					if (tag2 == '0')
						// it is the right center flag
						iFlag = FLAG_OUTSIDE_RC;
					else
					{
						// find the proper flag index
						switch (tag3)
						{
							case '5': posid++;
							case '4': posid++;
							case '3': posid++;
							case '2': posid++;
						}
						if (tag2 == 't')
							// it is a right top outside flag
							iFlag = FLAG_OUTSIDE_RT + posid;
						else
							// it is a right bottom outside flag
							iFlag = FLAG_OUTSIDE_RB + posid;
					}
					break;

				case 't':
					// it is one of the top flags of the field
					if (tag2 == '0')
						// it is the top center flag
						iFlag = FLAG_OUTSIDE_TC;
					else
					{
						// find the proper flag index
						switch (tag3)
						{
							case '5': posid++;
							case '4': posid++;
							case '3': posid++;
							case '2': posid++;
						}
						if (tag2 == 'l')
							// it is a top left outside flag
							iFlag = FLAG_OUTSIDE_TL + posid;
						else
							// it is a top right outside flag
							iFlag = FLAG_OUTSIDE_TR + posid;
					}
					break;

				case 'b':
					// it is one of the bottom flags of the field
					if (tag2 == '0')
						// it is the bottom center flag
						iFlag = FLAG_OUTSIDE_BC;
					else
					{
						// find the proper flag index
						switch (tag3)
						{
							case '5': posid++;
							case '4': posid++;
							case '3': posid++;
							case '2': posid++;
						}
						if (tag2 == 'l')
							// it is a bottom left outside flag
							iFlag = FLAG_OUTSIDE_BL + posid;
						else
							// it is a bottom right outside flag
							iFlag = FLAG_OUTSIDE_BR + posid;
					}
					break;
			}
			objInfo = m_flags[iFlag];


		} else if(m_token.startsWith("g")) {
			// we see a goal (currently G is not supported)
			m_token = m_tokenizer.nextToken();	// space
			if (m_token.equals(" ")) {
				m_token = m_tokenizer.nextToken();// side
				if (m_token.equals("l"))
					objInfo = m_leftGoal;
				else
					objInfo = m_rightGoal;
			}

		} else if(m_token.startsWith("b")) {
			objInfo = m_ball;
//			System.out.println("see ball");

		} else if(m_token.startsWith("l")) {
			m_token = m_tokenizer.nextToken();	// space
			m_token = m_tokenizer.nextToken();	// [l|r|t|b]
//			objInfo = new ObjectInfo(m_token.charAt(0));
		    objInfo = m_dummyLine;

		} else if(m_token.startsWith("P")) {
			objInfo = m_players[m_playerCount++];

		} else if(m_token.startsWith("B")) {
			objInfo = m_ball;

//		} else if(m_token.compareTo("G") == 0) {
//			objInfo = new ObjectInfo();

//		} else if(m_token.compareTo("F") == 0) {
//			objInfo = new ObjectInfo();

//		} else if(m_token.compareTo("L") == 0) {
//			objInfo = new ObjectInfo();
		}

		// Now we should find ")" m_token
		if(!m_token.equals(")"))
			m_token = m_tokenizer.nextToken();

		if (objInfo != null)
			return objInfo;
		else
			return m_dummyLine;
	}

	/**
	 * Checks if there is more information for this object
	 *
	 * @return	true if reached the end of this object, false otherwise
	 */
	private boolean findEnd()
	{
		try {
			m_token = m_tokenizer.nextToken();
			if(m_token.equals(")"))
			{
				m_token = m_tokenizer.nextToken();
				if(m_token.equals(" "))
					m_token = m_tokenizer.nextToken();
				return true;
			}
		} catch (NoSuchElementException e) {
			return true;
		}

		return false;
	}
}