/*
 * Copyright (c) 2000 living systems(R) AG, Germany. All rights reserved.
 * Author: Klaus Dorer
 *
 */

package com.ls.research.robocup.server;

import java.io.IOException;
import java.net.DatagramSocket;
import java.net.InetAddress;
import java.net.DatagramPacket;
import java.net.SocketException;
import java.net.UnknownHostException;



/**
 * This class represents the server on this client side
 *
 * @author  Klaus Dorer
 */

public class ServerRepresentation implements SoccerActionInterface
{
	// class memebers
	private DatagramSocket				m_socket;	// Socket to communicate with server
	private	InetAddress					m_host;		// Server address
	private int							m_port;		// server port
	private Receiver					m_receiver; // thread to receive messages

	/**
	 * Class to represent the soccerserver as if it was running locally.
	 * This class deals with the complete low level socket communication.
	 *
	 * @param player			reference to a SoccerPerceptionInterface used to notify the agent when
	 *							it receives a message from the server
	 * @param  host				host name of the server
	 * @param  port				port of the server
	 * @throws SocketException	if the Socket could not be initialized
	 * @throws UnknownHostException		if the host address is not valid
	 */
	 public ServerRepresentation(SoccerPerceptionInterface player, String host, int port)
	 throws SocketException, UnknownHostException
	 {
	 	m_socket 	= new DatagramSocket();
	 	m_host 		= InetAddress.getByName(host);
	 	m_port 		= port;
	 	m_receiver  = new Receiver(m_socket, player);
	 }

	 /**
	  * Closes the socket before object is deleted.
	  */
	 public void finalize()
	 {
		if (m_socket != null)
			disconnectFromServer();
	 		m_socket.close();
	 }

	//===========================================================================
	// Implementation of SoccerActionInterface

	 /**
	  * Moves the player to the specified position. Move is ignored by the server when the game is running
	  *
	  * @param x	x coordinate where to move to.
	  * @param y	y coordinate where to move to.
	  */
	 public void move(int x, int y)
	 {
	 	send("(" + ACTION_MOVE + " " + x + " " + y + ")");
	 }

	 /**
'	  * Turns the body of the player. The real direction change depends on the speed of the player.
	  * The head is of course also turned meaning that the player will look into a different direction
	  * after a turn. The headangle does not change.
	  *
	  * @param moment	the 'turn acceleration' with which to turn (in degree from -180 to 180).
	  */
	 public void turn(int moment)
	 {
	 	send("(" + ACTION_TURN + " " + moment + ")");
	 }

	 /**
	  * Turns the head of the player by the specified angle. Head turning is deterministic meaning that
	  * the head angle will in fact change by the specified amount but to a Max head angle of 90, min of
	  * -90 degrees.
	  *
	  * @param angle	the angle in degrees (-180 .. 180) by which to turn the head.
	  */
	 public void turnNeck(int angle)
	 {
	 	send("(" + ACTION_TURN_NECK + " " + angle + ")");
	 }

	 /**
	  * Accelerates the player by power into body direction. The player gets tired depending on power.
	  * When dashing backward (negative power) the player gets twice as tired.
	  *
	  * @param power	power of acceleration (-100 backwards to 100 forward)
	  */
	 public void dash(int power)
	 {
	 	send("(" + ACTION_DASH + " " + power + ")");
	 }

	 /**
	  * Kicks the ball with specified power into specified direction. The ball has to be within
	  * reach of the player. The actual power depends of the distance of the ball, and the relative
	  * direction to the body.
	  *
	  * @param power		the power with which to kick (1..100)
	  * @param direction	the direction (relative to body direction) in which to kick
	  */
	 public void kick(int power, int direction)
	 {
	 	send("(" + ACTION_KICK + " " + power + " " + direction + ")");
	 }

	 /**
	  * Command for the goalie to catch the ball. The ball has to be within a rectangular area
	  * (specified in server.conf) to be caught. The direction specifies the direction of the rectangle
	  *
	  * @param direction	direction in which to catch
	  */
	 public void catchBall(int direction)
	 {
	 	send("(" + ACTION_CATCH_BALL + " " + direction + ")");
	 }

	 /**
	  * The player says the string passed. Note that only one player of a team is allowed to say
	  * something every second cycle. The length of the message is restriced (see server.conf)
	  *
	  * @param message	message that is said by the player
	  */
	 public void say(String message)
	 {
	 	send("(" + ACTION_SAY + " " + message + ")");
	 }

	 /**
	  * Allows to change the view mode of the player.
	  *
	  * @param 	angle	narrow	for 45 degree view every 75 ms,
	  *					normal	for 90 degree view every 150ms (default)
	  *					wide	for 180 degree view every 300ms.
	  * @param	quality	low		for low quality but high speed information
	  *					high	for high quality but lower speed information (default)
	  */
	 public void changeView(String angle, String quality)
	 {
	 	send("(" + ACTION_CHANGE_VIEW + " " + angle + " " + quality + ")");
	 }

	 /**
	  * Player wants to receive additional body information. Can only be done once a cycle. Since
	  * server version 5 this is automatically sent every 100 ms and therefore rarely used.
	  */
	 public void senseBody()
	 {
	 	send("(" + ACTION_SENSE_BODY + ")");
	 }

	 /**
	  * Player wants to receive the score of the game.
	  */
	 public void score()
	 {
	 	send("(" + ACTION_SCORE + ")");
	 }

	 /**
	  * Called to connect the player to the soccer server
	  *
	  * @param team		name of the team to join
	  * @param version	protocol version to use e.g. "version 6.00" for server version 6
	  * @param goalie	flag indicating if this is the goalie
	  */
	 public void connectToServer(String team, String version, boolean goalie)
	 {
	 	// start receiver thread
	 	if (goalie)
	 		send("(" + ACTION_CONNECT + " " + team + " ("+ CONNECT_GOALIE + ") (" + version + "))");
	 	else
	 		send("(" + ACTION_CONNECT + " " + team + " (" + version + "))");

		m_receiver.start();
	 }

	 /**
	  * Reconnects a player to the server as the player specified.
	  * This is used to disconnect the player in half time and replace it with a different one.
	  *
	  * @param team		name of the team to join
	  * @param number	number of the player that this player wants to take over
	  */
	 public void reConnectToServer(String team, int number)
	 {
	 	// start receiver thread
		send("(" + ACTION_RECONNECT + " " + team + " " + number);
		m_receiver.start();
	 }

	 /**
	  * Disconnects the player from the server
	  */
	 public void disconnectFromServer()
	 {
	 	// stop receiver thread
	 	m_receiver.stopThread();
	 	// make sure we get one more message
		senseBody();
		send("(" + ACTION_DISCONNECT + ")");
	 }


	 /**
	  * Sends the passed string accross the socket to the server
	  *
	  * @param message	the message to send
	  */
	 private void send(String message)
	 {
	 	byte[] buffer = message.getBytes();
	 	DatagramPacket packet = new DatagramPacket(buffer, buffer.length, m_host, m_port);

	 	try{
	 		m_socket.send(packet);
	 	} catch(IOException e) {
	 		System.err.println("socket sending error " + e);
	 	}
	 }

	/**
	 * Class dealing with the concurrent reception of server information
	 */
	class Receiver extends Thread
	{
		private SoccerPerceptionInterface	m_player;	// notification for player
		private DatagramSocket				m_socket;
		private boolean						m_isRunning;

		/**
		 * Constructor receiving the socket on which to listen and the object which should be
		 *
		 * @param socket	datagram socket on which to listen
		 */
		public Receiver(DatagramSocket socket, SoccerPerceptionInterface player)
		{
			m_socket = socket;
			m_player = player;
			m_isRunning = false;
		}

		/**
		 * Reads from the socket until someone stops the thread.
		 * If something is received the player is notified
		 */
		public void run()
		{
			String message;

			m_isRunning = true;
			while (m_isRunning) {
				message = receive();

				// Notify depending on kind of information
				if(message.startsWith("(sense")) {
					m_player.onSenseBody(message);

				} else if(message.startsWith("(see")) {
					m_player.onSee(message);

				} else if(message.startsWith("(hear")) {
					m_player.onHear(message);

				} else if(message.startsWith("(change")) {
					m_player.onChangePlayer(message);

				} else if(message.startsWith("(init")) {
					m_player.onInit(message);

				} else if(message.startsWith("(server_")) {
					m_player.onServerParam(message);

				} else if(message.startsWith("(player_param")) {
					m_player.onPlayerParam(message);

				} else if(message.startsWith("(player_type")) {
					m_player.onPlayerType(message);

				} else if(message.startsWith("(score")) {
					m_player.onScore(message);

				} else {
					m_player.onMessage(message);
				}
			}
		}

		/**
		 * Stops this thread after receiving the next information.
		 */
		public void stopThread()
		{
			m_isRunning = false;
		}

		 /**
		  * Waits to receive a message from the server
		  *
		  * @return 	the message	received by the server
		  */
		private String receive()
		{
			final int MSG_SIZE = 1024;
			byte[] buffer = new byte[MSG_SIZE];
			DatagramPacket packet = new DatagramPacket(buffer, MSG_SIZE);

			try{
				m_socket.receive(packet);
				m_port = packet.getPort();
			} catch(IOException e) {
				System.err.println("socket receiving error " + e);
			}
			return new String(buffer, 0, packet.getLength());
		}
	}
}
